from typing import Any, Literal, List

from ...action import RemoveObjectAction

from ...event import MakeDamageEventArguments

from ...consts import ElementType, IconType, SkillType

from ...modifiable_values import DamageIncreaseValue
from .base import RoundCharactorStatus


class HeavyStrike(RoundCharactorStatus):
    name: Literal['Heavy Strike'] = 'Heavy Strike'
    desc: str = (
        "During this round, your current active character's next "
        'Normal Attack deals +1 DMG. '
        'When this Normal Attack is a Charged Attack: Deal +1 additional DMG.'
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        Increase damage for normal attack by 1 and decrease usage. 
        If it is charged attack, increase damage more 1.
        """
        assert mode == 'REAL'
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not this charactor use normal attack, not modify
            return value
        if self.usage <= 0:
            # no usage, not modify
            return value
        # this charactor use normal attack, increase damage by 1
        value.damage += 1
        # if it is charged attack, increase damage more 1
        if match.player_tables[self.position.player_idx].charge_satisfied:
            value.damage += 1
        # decrease usage
        self.usage -= 1
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()


class ShatteringIce(RoundCharactorStatus):
    name: Literal[
        'Elemental Resonance: Shattering Ice'
    ] = 'Elemental Resonance: Shattering Ice'
    desc: str = (
        'During this Round, your character will deal +2 DMG '
        'for the next instance.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        add 2 damage for skills.
        """
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not this charactor use skill, do nothing
            return value
        if self.usage <= 0:
            # no usage, do nothing
            return value
        # we trigger elemental reaction, add 2 damage
        assert mode == 'REAL'
        self.usage -= 1
        value.damage += 2
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When make damage end, check whether to remove.
        """
        return self.check_should_remove()


class FerventFlames(RoundCharactorStatus):
    name: Literal[
        'Elemental Resonance: Fervent Flames'
    ] = 'Elemental Resonance: Fervent Flames'
    desc: str = (
        'During this round, the next instance of Pyro-Related Reactions your '
        'character triggers deals +3 DMG.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        If we trigger pyro elemental reaction, add 3 damage.
        """
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not this charactor use skill, do nothing
            return value
        if ElementType.PYRO not in value.reacted_elements:
            # not trigger pyro elemental reaction, do nothing
            return value
        if self.usage <= 0:
            # no usage, do nothing
            return value
        # we trigger elemental reaction, add 2 damage
        assert mode == 'REAL'
        self.usage -= 1
        value.damage += 3
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When make damage end, check whether to remove.
        """
        return self.check_should_remove()


EventCardCharactorStatus = HeavyStrike | ShatteringIce | FerventFlames



from .nahida import Nahida, TheSeedOfStoredKnowledge
from .collei import Collei, FloralSidewinder, CuileinAnbar
from .tighnari import Tighnari, KeenSight, ClusterbloomArrow
from .jadeplume_terrorshroom import JadeplumeTerrorshroom, ProliferatingSpores
from .yaoyao import Yaoyao, YueguiThrowingMode, Beneficent


DendroCharactors = (
    Collei | Tighnari | Nahida | Yaoyao
    # finally monsters
    | JadeplumeTerrorshroom
)
SummonsOfDendroCharactors = (
    CuileinAnbar | ClusterbloomArrow | YueguiThrowingMode
)
DendroCharactorTalents = (
    FloralSidewinder | KeenSight | TheSeedOfStoredKnowledge | Beneficent
    # finally monsters
    | ProliferatingSpores
)



from typing import Any, List, Literal

from ....consts import DamageType, ObjectPositionType

from ....action import Actions

from ....event import MakeDamageEventArguments

from ....modifiable_values import DamageIncreaseValue
from .base import ArtifactBase
from ....struct import Cost


class CrownOfWatatsumi(ArtifactBase):
    name: Literal['Crown of Watatsumi']
    desc: str = (
        'For every 3 HP of healing your characters receive, this card '
        'accumulates 1 Sea-Dyed Foam (maximum of 2). '
        'When this character deals DMG: Consume all Sea-Dyed Foam. DMG is '
        'increased by 1 for each Sea-Dyed Foam consumed.'
    )
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost(same_dice_number = 1)
    usage: int = 0
    max_usage: int = 2
    counter: int = 0

    def equip(self, match: Any) -> List[Actions]:
        self.usage = 0
        self.counter = 0
        return []

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, 
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not self use damage skill
            return value
        # increase damage
        assert mode == 'REAL'
        value.damage += self.usage
        self.usage = 0
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        If our player receives heal, add counter.
        """
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return []
        for damage in event.damages:
            if (
                damage.final_damage.target_position.player_idx
                != self.position.player_idx
                or damage.final_damage.damage_type != DamageType.HEAL
            ):
                # target not self player, or not heal
                continue
            # increase counter by heal
            self.counter += damage.hp_after - damage.hp_before
            # self.counter += - damage.final_damage.damage
            self.usage += self.counter // 3
            self.counter = self.counter % 3
            if self.usage >= self.max_usage:
                # usage full
                self.counter = 0
                self.usage = self.max_usage
        return []


OceanHuedArtifacts = CrownOfWatatsumi | CrownOfWatatsumi




from typing import Any, List, Literal

from ....action import Actions, ChargeAction

from ....action import ChangeObjectUsageAction, CreateObjectAction
from ....event import (
    ChargeEventArguments, RoundPrepareEventArguments, SkillEndEventArguments
)

from ....modifiable_values import DamageIncreaseValue
from ....status.charactor_status.base import ShieldCharactorStatus
from ....status.team_status.base import ShieldTeamStatus

from ....consts import FactionType, ObjectPositionType, SkillType, WeaponType

from ....struct import Cost
from .base import RoundEffectWeaponBase, WeaponBase


class VortexVanquisher(RoundEffectWeaponBase):
    name: Literal['Vortex Vanquisher']
    desc: str = (
        'The character deals +1 DMG. When your active character is protected '
        'by a Shield Character Status or a Shield Combat Status, you deal +1 '
        'DMG. After the character uses an Elemental Skill: If you have a '
        'Combat Status that grants a Shield on your side, add 1 Shield point '
        'to that Combat Status. (Once per Round)'
    )
    cost: Cost = Cost(same_dice_number = 3)
    version: Literal['3.7'] = '3.7'
    weapon_type: WeaponType = WeaponType.POLEARM
    max_usage_per_round: int = 1

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, 
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        If self has shield, +1 damage.
        """
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return value
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not current charactor using skill
            return value
        # modify damage
        assert mode == 'REAL'
        value.damage += self.damage_increase
        # if have shield, increase one more
        have_shield = False
        team_status = match.player_tables[self.position.player_idx].team_status
        for status in team_status:
            if issubclass(type(status), ShieldTeamStatus):
                have_shield = True
        if not have_shield:
            # no shield team status, check charactor status
            charactor = match.player_tables[
                self.position.player_idx].charactors[
                    self.position.charactor_idx]
            for status in charactor.status:
                if issubclass(type(status), ShieldCharactorStatus):
                    have_shield = True
        if have_shield:
            value.damage += 1
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        if self charactor use elemental skill, check whether have
        shield team status. If so, add 1 usage. 
        """
        if event.action.skill_type != SkillType.ELEMENTAL_SKILL:
            # not elemental skill
            return []
        if self.usage <= 0:
            # no usage
            return []
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
            source_area = ObjectPositionType.CHARACTOR
        ):
            # not self charactor use skill or not equipped
            return []
        team_status = match.player_tables[self.position.player_idx].team_status
        for status in team_status:
            if issubclass(type(status), ShieldTeamStatus):
                # find shield status, add 3 usage
                self.usage -= 1
                return [ChangeObjectUsageAction(
                    object_position = status.position,
                    change_type = 'DELTA',
                    change_usage = 1,
                )]
        return []


class LithicSpear(WeaponBase):
    name: Literal['Lithic Spear']
    desc: str = (
        'The character deals +1 DMG. '
        'When played: For each party member from Liyue, grant 1 Shield point '
        'to the character to which this is attached. (Max 3 points)'
    )
    cost: Cost = Cost(same_dice_number = 3)
    version: Literal['3.7'] = '3.7'
    weapon_type: WeaponType = WeaponType.POLEARM

    def generate_shield(self, count: int) -> List[Actions]:
        count = min(count, 3)
        if count > 0:
            return [CreateObjectAction(
                object_position = self.position.set_area(
                    ObjectPositionType.CHARACTOR_STATUS),
                object_name = self.name,
                object_arguments = {
                    'usage': count,
                    'max_usage': count,
                }
            )]
        return []

    def equip(self, match: Any) -> List[Actions]:
        """
        Generate shield
        """
        count = 0
        for c in match.player_tables[self.position.player_idx].charactors:
            if FactionType.LIYUE in c.faction:
                count += 1
        return self.generate_shield(count)


class EngulfingLightning(RoundEffectWeaponBase):
    name: Literal['Engulfing Lightning']
    desc: str = (
        'The character deals +1 DMG. '
        'Triggers automatically once per Round: If the character to which '
        'this is attached does not have Energy, they will gain 1 Energy.'
    )
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 3)
    weapon_type: WeaponType = WeaponType.POLEARM
    max_usage_per_round: int = 1

    def _charge_one(self, match: Any) -> List[ChargeAction]:
        if self.position.area != ObjectPositionType.CHARACTOR:
            # not equipped
            return []
        charactor = match.player_tables[
            self.position.player_idx].charactors[self.position.charactor_idx]
        if charactor.charge == 0 and self.usage > 0:
            self.usage -= 1
            return [ChargeAction(
                player_idx = self.position.player_idx,
                charactor_idx = self.position.charactor_idx,
                charge = 1,
            )]
        return []

    def equip(self, match: Any) -> List[Actions]:
        return super().equip(match) + self._charge_one(match)

    def event_handler_CHARGE(
        self, event: ChargeEventArguments, match: Any
    ) -> List[ChargeAction]:
        return self._charge_one(match)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        ret = super().event_handler_ROUND_PREPARE(event, match)
        return ret + self._charge_one(match)


Polearms = LithicSpear | VortexVanquisher | EngulfingLightning



from typing import Any, List, Literal

from ...consts import ELEMENT_TO_DIE_COLOR

from ...action import (
    ActionTypes, ChargeAction, CreateDiceAction, CreateObjectAction, 
    RemoveObjectAction
)
from ...struct import Cost, ObjectPosition
from .others import IHaventLostYet as IHLY_4_0
from .others import SendOff as SendOff_3_7
from .foods import MintyMeatRolls as MintyMeatRolls_3_4
from .resonance import ThunderAndEternity as TAE_4_0
from .others import MasterOfWeaponry as MOW_4_1
from .others import BlessingOfTheDivineRelicsInstallation as BOTDRI_4_1
from .resonance import WindAndFreedom as WAF_4_1
from .foods import TeyvatFriedEgg as TFE_4_1


class ThunderAndEternity_3_7(TAE_4_0):
    name: Literal['Thunder and Eternity']
    desc: str = (
        'Convert all your Elemental Dice to the Type of the active character. '
    )
    version: Literal['3.7']

    def get_dice_color(self, match: Any) -> str:
        """
        Convert all your Elemental Dice to the Type of the active character.
        """
        charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        return ELEMENT_TO_DIE_COLOR[charactor.element]


class IHaventLostYet_3_3(IHLY_4_0):
    name: Literal["I Haven't Lost Yet!"]
    desc: str = (
        "Only playable if one of your characters is defeated this Round: "
        "Create Omni Element x1 and your current active character gains 1 "
        "Energy. "
    )
    version: Literal['3.3']
    cost: Cost = Cost()

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ChargeAction | CreateDiceAction | CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert ret[-1].type == ActionTypes.CREATE_OBJECT
        # old version not adding team status
        ret = ret[:-1]
        return ret


class SendOff_3_3(SendOff_3_7):
    name: Literal['Send Off']
    desc: str = '''Choose one Summon on the opposing side and destroy it.'''
    version: Literal['3.3']
    cost: Cost = Cost(any_dice_number = 2)

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[RemoveObjectAction]:
        """
        Act the card. Create team status.
        """
        assert target is not None
        return [RemoveObjectAction(
            object_position = target,
        )]


class MintyMeatRolls_3_3(MintyMeatRolls_3_4):
    version: Literal['3.3']
    desc: str = (
        "Before this Round ends, the target character's Normal Attacks cost 1 "
        "less Unaligned Element."
    )


class MasterOfWeaponry_3_3(MOW_4_1):
    desc: str = (
        'Shift 1 Weapon Equipment Card that has been equipped to one of your '
        'characters to another one of your characters of the same Weapon '
        'Type.'
    )
    version: Literal['3.3']
    reset_usage: bool = False


class BlessingOfTheDivineRelicsInstallation_3_3(BOTDRI_4_1):
    desc: str = (
        'Shift 1 Artifact Equipment Card that has been equipped to one of '
        'your characters to another one of your characters. '
    )
    version: Literal['3.3']
    reset_usage: bool = False


class WindAndFreedom_3_7(WAF_4_1):
    name: Literal['Wind and Freedom']
    desc: str = (
        'In this Round, when an opposing character is defeated during your '
        'Action, you can continue to act again when that Action ends. '
    )
    version: Literal['3.7']


class TeyvatFriedEgg_3_7(TFE_4_1):
    version: Literal['3.7'] = '3.7'
    cost: Cost = Cost(same_dice_number = 3)


OldVersionEventCards = (
    WindAndFreedom_3_7 | ThunderAndEternity_3_7 | TeyvatFriedEgg_3_7

    | IHaventLostYet_3_3 | MasterOfWeaponry_3_3 
    | BlessingOfTheDivineRelicsInstallation_3_3 | SendOff_3_3 
    | MintyMeatRolls_3_3
)



from typing import List, Literal

from ..pyro.xiangling import Xiangling as X_3_8
from ..pyro.xiangling import Pyronado as P_3_8
from ..pyro.xiangling import DoughFu, GuobaAttack


class Pyronado(P_3_8):
    desc: str = '''Deals 2 Pyro DMG, creates 1 Pyronado.'''
    damage: int = 2


class Xiangling_3_3(X_3_8):
    version: Literal['3.3']
    skills: List[DoughFu | GuobaAttack | Pyronado] = []

    def _init_skills(self) -> None:
        self.skills = [
            DoughFu(),
            GuobaAttack(),
            Pyronado()
        ]



import numpy as np
from typing import List, Any
from src.lpsim.agents.interaction_agent import (
    InteractionAgent, InteractionAgent_V1_0
)
from src.lpsim.server.match import Match
from src.lpsim.server.event_handler import OmnipotentGuideEventHandler
from src.lpsim.agents.agent_base import AgentBase
from src.lpsim.server.struct import ObjectPosition
from src.lpsim.utils import BaseModel
from tests.default_random_state import get_default_random_state


def set_16_omni(match: Match):
    match.config.initial_dice_number = 16
    match.event_handlers.append(OmnipotentGuideEventHandler())


def make_respond(agent: AgentBase, match: Match, assertion: bool = True):
    """
    Make response once, and step the match until new request or
    game end, regardless of the target of requests.
    If not requesting this agent and nees assetion, raise AssertionError.
    Otherwise, do nothing.
    """
    resp = agent.generate_response(match)
    if assertion:
        assert resp is not None
    if resp is None:
        return
    match.respond(resp)
    match.step()


def check_hp(match: Match, hp: List[List[int]]):
    """
    input hps for each player and check
    """
    hps = [[x.hp for x in t.charactors] for t in match.player_tables]
    assert hps == hp


def check_name(name: str, lists: List[Any], exist: bool = True):
    """
    check the existence of name in lists
    """
    real_exist = name in [x.name for x in lists]
    assert real_exist == exist


def remove_ids(model: BaseModel) -> BaseModel:
    """
    remove ids of all objects in-place.
    """
    for key in model.__fields__.keys():
        value = getattr(model, key)
        if isinstance(value, ObjectPosition):
            model.__setattr__(key, value.set_id(0))
        elif isinstance(value, BaseModel):
            remove_ids(value)
        elif isinstance(value, list) or isinstance(value, tuple):
            assert isinstance(value, list)
            for num, v in enumerate(value):
                if isinstance(v, ObjectPosition):
                    value[num] = v.set_id(0)
                elif isinstance(v, BaseModel):
                    remove_ids(v)
    if 'id' in model.__fields__.keys():
        model.id = 0
    return model


def get_random_state(offset: int = 0):
    """
    get random state tuple for test. it will use a fixed random state and
    random offset times as the result random state.
    """
    state: List = get_default_random_state()
    if offset > 0:
        random_state = np.random.RandomState()
        state[1] = np.array(state[1], dtype = 'uint32')
        random_state.set_state(state)  # type: ignore
        for _ in range(offset):
            random_state.random()
        state = list(random_state.get_state(legacy = True))  # type: ignore
        state[1] = state[1].tolist()
    return state


def get_test_id_from_command(agent: InteractionAgent | InteractionAgent_V1_0):
    """
    from commands to get test number. return test number and remove this
    test message from commands if success. otherwise return 0.
    A command should like: `TEST 1 other information about this test`,
    and will return 1.
    """
    if len(agent.commands) == 0:  # pragma: no cover
        return 0
    command = agent.commands[0]
    if command.startswith('TEST'):
        agent.commands = agent.commands[1:]
        return int(command.split(' ')[1])
    return 0


def enable_logging():  # pragma: no cover
    import logging
    logging.basicConfig(level = logging.INFO)


def get_pidx_cidx(cmd):
    return int(cmd[2][1]), int(cmd[2][3])


def check_usage(data, usage_strs):
    usage = [int(x) for x in usage_strs]
    assert len(data) == len(usage)
    for u, d in zip(usage, data):
        assert u == d.usage



from typing import Any, List, Literal

from ....consts import FactionType

from ....action import Actions

from .other_polearm import LithicSpear as LS_3_7


class LithicSpear_3_3(LS_3_7):
    desc: str = (
        'The character deals +1 DMG. '
        'When played: For each alive party member from Liyue, grant 1 Shield '
        'point to the character to which this is attached. (Max 3 points)'
    )
    version: Literal['3.3']

    def equip(self, match: Any) -> List[Actions]:
        """
        Generate shield
        """
        count = 0
        for c in match.player_tables[self.position.player_idx].charactors:
            if c.is_alive and FactionType.LIYUE in c.faction:
                count += 1
        return self.generate_shield(count)


OldVersionWeapons = LithicSpear_3_3 | LithicSpear_3_3



from typing import Literal

from ...struct import Cost

from ...consts import DieColor
from ..cryo.diona import ShakenNotPurred as SNP_4_1


class ShakenNotPurred_3_3(SNP_4_1):
    version: Literal['3.3']
    cost: Cost = Cost(
        elemental_dice_color = DieColor.CRYO,
        elemental_dice_number = 4
    )





from typing import Any, Literal, List

from ...struct import Cost

from ...consts import (
    CostLabels, DamageElementalType, DamageType, DieColor, ElementType, 
    ElementalReactionType, IconType, ObjectPositionType, PlayerActionLabels
)

from ...action import (
    Actions, ChangeObjectUsageAction, CreateDiceAction, DrawCardAction, 
    MakeDamageAction, RemoveObjectAction, SwitchCharactorAction
)

from ...event import (
    CharactorDefeatedEventArguments, ActionEndEventArguments, 
    DeclareRoundEndEventArguments, 
    MakeDamageEventArguments, MoveObjectEventArguments, 
    RoundPrepareEventArguments, SkillEndEventArguments
)

from ...modifiable_values import (
    CombatActionValue, CostValue, DamageIncreaseValue, DamageValue
)
from .base import (
    RoundTeamStatus, ShieldTeamStatus, TeamStatusBase, UsageTeamStatus
)


class ChangingShifts(UsageTeamStatus):
    name: Literal['Changing Shifts'] = 'Changing Shifts'
    desc: str = (
        'The next time you perform "Switch Character": '
        'Spend 1 less Elemental Die.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def value_modifier_COST(
        self, value: CostValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        assert self.usage > 0
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
        ):
            # not self switch charactor, do nothing
            return value
        if not (value.cost.label & CostLabels.SWITCH_CHARACTOR.value):
            # not switch charactor, do nothing
            return value
        # decrease 1 any cost
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == 'REAL':
                self.usage -= 1
        return value

    def event_handler_SWITCH_CHARACTOR(
        self, event: ActionEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When switch charactor end, check whether to remove.
        """
        return self.check_should_remove()


class IHaventLostYet(RoundTeamStatus):
    name: Literal["I Haven't Lost Yet!"] = "I Haven't Lost Yet!"
    desc: str = '''You cannot play "I Haven't Lost Yet!" again this round.'''
    version: Literal['4.0'] = '4.0'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.DEBUFF] = IconType.DEBUFF


class FreshWindOfFreedom(RoundTeamStatus):
    name: Literal['Fresh Wind of Freedom'] = 'Fresh Wind of Freedom'
    desc: str = (
        'In this Round, when an opposing character is defeated during your '
        'Action, you can continue to act again when that Action ends. '
        'Usage(s): 1 '
    )
    version: Literal['4.1'] = '4.1'
    usage: int = 1
    max_usage: int = 1
    activated: bool = False
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def event_handler_CHARACTOR_DEFEATED(
        self, event: CharactorDefeatedEventArguments, match: Any
    ) -> List[Actions]:
        """
        When an enemy charactor is defeated, mark activated.
        """
        if (self.position.player_idx != event.action.player_idx):
            # enemy defeated, mark activated.
            self.activated = True
        return []

    def value_modifier_COMBAT_ACTION(
        self, value: CombatActionValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> CombatActionValue:
        """
        When enemy charactor defeated, 
        """
        assert mode == 'REAL'
        if not self.activated:
            # not activated, do nothing
            return value
        assert self.usage > 0
        assert self.position.check_position_valid(
            value.position, match, player_idx_same = True,
        )
        # combat action end from self, if combat action, change to quick.
        if value.do_combat_action:
            value.do_combat_action = False
            self.usage -= 1
        # else:
        #     # Mona + Kaeya may trigger
        return value

    def event_handler_ACTION_END(
        self, event: ActionEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When action end event, check whether to remove.
        """
        return self.check_should_remove()


class LeaveItToMe(UsageTeamStatus):
    name: Literal['Leave It to Me!']
    desc: str = (
        'The next time you perform "Switch Character": '
        'The switch will be considered a Fast Action instead of a '
        'Combat Action.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def value_modifier_COMBAT_ACTION(
        self, value: CombatActionValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> CombatActionValue:
        assert self.usage > 0
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
        ):
            # not self switch charactor, do nothing
            return value
        if value.action_label & PlayerActionLabels.SWITCH.value == 0:
            # not switch charactor, do nothing
            return value
        if not value.do_combat_action:
            # already quick action, do nothing
            return value
        # self switch charactor, change to quick action
        value.do_combat_action = False
        assert mode == 'REAL'
        self.usage -= 1
        return value

    def event_handler_ACTION_END(
        self, event: ActionEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When action end event, check whether to remove.
        """
        return self.check_should_remove()


class EnduringRock(RoundTeamStatus):
    """
    Made Geo damage is determined by the following:
    in value_modifier_DAMAGE_INCREASE, check whether it is caused by our 
    charactor and is Geo damage. If so, mark did_geo_attack.
    in event_handler_SKILL_END, check whether did_geo_attack is True. If so,
    trigger the following check. And regardless of the result, reset
    did_geo_attack to False.

    As DAMAGE_INCREASE with skill as source must have SKILL_END in the 
    following, it will not mix other Geo damage or other charactor's skills.

    TODO: with Sparks 'n' Splash
    """
    name: Literal[
        'Elemental Resonance: Enduring Rock'
    ] = 'Elemental Resonance: Enduring Rock'
    desc: str = (
        'During this round, after your character deals Geo DMG next time: '
        'Should there be any Combat Status on your side that provides Shield, '
        'grant one such Status with 3 Shield points.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    did_geo_attack: bool = False
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not our charactor use skill, do nothing
            return value
        if value.damage_elemental_type != ElementType.GEO:
            # not Geo damage, do nothing
            return value
        # our charactor use Geo damage skill, mark did_geo_attack
        self.did_geo_attack = True
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction | RemoveObjectAction]:
        """
        if self charactor use skill, and did geo attack, check whether have
        shield team status. If so, add 3 usage. Regardless of the result, reset
        did_geo_attack to False.
        """
        if not self.did_geo_attack:
            # not fit condition, do nothing
            return []
        # reset did_geo_attack
        self.did_geo_attack = False
        if (
            self.position.player_idx != event.action.position.player_idx
        ):  # pragma: no cover
            # not self charactor use skill, do nothing
            return []
        team_status = match.player_tables[self.position.player_idx].team_status
        for status in team_status:
            if issubclass(type(status), ShieldTeamStatus):
                # find shield status, add 3 usage
                self.usage -= 1
                return [ChangeObjectUsageAction(
                    object_position = status.position,
                    change_type = 'DELTA',
                    change_usage = 3,
                )] + self.check_should_remove()
        return list(self.check_should_remove())


class WhereIstheUnseenRazor(RoundTeamStatus):
    name: Literal['Where Is the Unseen Razor?'] = 'Where Is the Unseen Razor?'
    desc: str = (
        'During this Round, the next time you play a Weapon card: '
        'Spend 2 less Elemental Dice.'
    )
    version: Literal['4.0'] = '4.0'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        decrease weapons cost by 2
        """
        if not value.cost.label & CostLabels.WEAPON.value:
            # not weapon, do nothing
            return value
        # try decrease twice
        success = [value.cost.decrease_cost(None), 
                   value.cost.decrease_cost(None)]
        if True in success:  # pragma: no branch
            # decrease success at least once
            if mode == 'REAL':
                self.usage -= 1
        return value

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        As it triggers on equipping weapon,
        When move object end, check whether to remove.
        """
        return self.check_should_remove()


class SprawlingGreenery(RoundTeamStatus):
    name: Literal[
        'Elemental Resonance: Sprawling Greenery'
    ] = 'Elemental Resonance: Sprawling Greenery'
    desc: str = (
        'During this round, the next Elemental Reaction you trigger deals '
        '+2 DMG.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        If we trigger elemental reaction, add 2 damage.
        """
        if (
            self.position.player_idx != value.position.player_idx
            or self.position.player_idx == value.target_position.player_idx
        ):
            # not we attack or we attack ourself, do nothing
            return value
        if value.damage_type != DamageType.DAMAGE:  # pragma: no cover
            # not damage, do nothing
            return value
        if value.element_reaction == ElementalReactionType.NONE:
            # not elemental reaction, do nothing
            return value
        if self.usage <= 0:
            # no usage, do nothing
            return value
        # we trigger elemental reaction, add 2 damage
        assert mode == 'REAL'
        self.usage -= 1
        value.damage += 2
        return value

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When make damage end, check whether to remove.
        """
        return self.check_should_remove()


class ReviveOnCooldown(RoundTeamStatus):
    name: Literal['Revive on cooldown'] = 'Revive on cooldown'
    desc: str = '''You cannot revive any charactor with food this round.'''
    version: Literal['3.7'] = '3.7'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.FOOD] = IconType.FOOD


class StoneAndContracts(TeamStatusBase):
    name: Literal['Stone and Contracts']
    desc: str = (
        'When the Action Phase of the next Round begins: Create 3 Omni '
        'Element. '
    )
    version: Literal['3.7'] = '3.7'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def event_handler_ROUND_PREPARE(
        self, event: Any, match: Any
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        When round prepare, create 3 omni element and remove self.
        """
        return [
            CreateDiceAction(
                player_idx = self.position.player_idx,
                color = DieColor.OMNI,
                number = 3,
            ),
            RemoveObjectAction(
                object_position = self.position,
            )
        ]


class AncientCourtyard(RoundTeamStatus):
    name: Literal['Ancient Courtyard']
    desc: str = (
        'You must have a character who has already equipped a Weapon or '
        'Artifact: The next time you play a Weapon or Artifact card in this '
        'Round: Spend 2 less Elemental Dice.'
    )
    version: Literal['3.8'] = '3.8'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        decrease weapons or artifact cost by 2
        """
        if value.position.player_idx != self.position.player_idx:
            # not self charactor, do nothing
            return value
        if value.cost.label & (CostLabels.WEAPON.value 
                               | CostLabels.ARTIFACT.value) == 0:
            # not weapon or artifact, do nothing
            return value
        # try decrease twice
        success = [value.cost.decrease_cost(None), 
                   value.cost.decrease_cost(None)]
        if True in success:
            # decrease success at least once
            if mode == 'REAL':
                self.usage -= 1
        return value

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        As it triggers on equipping weapon or artifact,
        When move object end, check whether to remove.
        """
        return self.check_should_remove()


class FatuiAmbusher(UsageTeamStatus):
    name: Literal[
        'Fatui Ambusher: Cryo Cicin Mage',
        'Fatui Ambusher: Mirror Maiden',
        'Fatui Ambusher: Pyroslinger Bracer',
        'Fatui Ambusher: Electrohammer Vanguard'
    ]
    desc: str = (
        "After a character on whose side of the field this card is on uses a "
        "Skill: Deals 1 XXX DMG to the active character on that side. "
        "(Once per round) "
        "Usage(s): 2"
    )
    element: DamageElementalType = DamageElementalType.PIERCING
    version: Literal['3.7'] = '3.7'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    activated_this_round: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.name == 'Fatui Ambusher: Cryo Cicin Mage':
            self.element = DamageElementalType.CRYO
        elif self.name == 'Fatui Ambusher: Mirror Maiden':
            self.element = DamageElementalType.HYDRO
        elif self.name == 'Fatui Ambusher: Pyroslinger Bracer':
            self.element = DamageElementalType.PYRO
        else:
            assert self.name == 'Fatui Ambusher: Electrohammer Vanguard'
            self.element = DamageElementalType.ELECTRO
        self.desc = self.desc.replace('XXX', self.element.name.capitalize())

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        """
        reset activated_this_round
        """
        self.activated_this_round = False
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        When attached charactor use skill, then damage itself.
        """
        if self.activated_this_round:
            # already activated, do nothing
            return []
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True, 
            target_area = ObjectPositionType.SKILL,
        ):
            # not charactor use skill, not modify
            return []
        active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        if self.usage > 0:  # pragma: no branch
            self.usage -= 1
            self.activated_this_round = True
            return [MakeDamageAction(
                source_player_idx = self.position.player_idx,
                target_player_idx = self.position.player_idx,
                damage_value_list = [DamageValue(
                    position = self.position,
                    damage_type = DamageType.DAMAGE,
                    target_position = active_charactor.position,
                    damage = 1,
                    damage_elemental_type = self.element,
                    cost = Cost()
                )]
            )]
        else:
            return []  # pragma: no cover


class RhythmOfTheGreatDream(UsageTeamStatus):
    name: Literal['Rhythm of the Great Dream']
    desc: str = (
        'The next time you play a Weapon or Artifact from your hand: Spend 1 '
        'less Elemental Die.'
    )
    version: Literal['3.8'] = '3.8'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.BUFF] = IconType.BUFF

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal['TEST', 'REAL']
    ) -> CostValue:
        """
        decrease weapons or artifact cost by 1
        """
        if value.position.player_idx != self.position.player_idx:
            # not self charactor, do nothing
            return value
        if value.cost.label & (CostLabels.WEAPON.value 
                               | CostLabels.ARTIFACT.value) == 0:
            # not weapon or artifact, do nothing
            return value
        # try decrease once
        if value.cost.decrease_cost(None):
            # decrease success
            if mode == 'REAL':
                self.usage -= 1
        return value

    def event_handler_MOVE_OBJECT(
        self, event: MoveObjectEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        As it triggers on equipping weapon or artifact,
        When move object end, check whether to remove.
        """
        return self.check_should_remove()


class WhenTheCraneReturned(UsageTeamStatus):
    name: Literal['When the Crane Returned']
    desc: str = (
        'The next time you use a Skill: Switch your next character in to be '
        'the active character.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    decrease_usage_when_trigger: bool = True
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[SwitchCharactorAction]:
        """
        when our charactor used skill, switch to next charactor
        """
        if self.position.player_idx != event.action.position.player_idx:
            # not our charactor, do nothing
            return []
        if self.usage <= 0:  # pragma: no cover
            return []
        next_idx = match.player_tables[
            self.position.player_idx].next_charactor_idx()
        if next_idx is None:
            # no next charactor, do nothing
            return []
        if self.decrease_usage_when_trigger:
            self.usage -= 1
        return [SwitchCharactorAction(
            player_idx = self.position.player_idx,
            charactor_idx = next_idx,
        )]

    def event_handler_SWITCH_CHARACTOR(
        self, event: ActionEndEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When switch charactor end, check whether to remove.
        """
        return self.check_should_remove()


class WindAndFreedom(WhenTheCraneReturned, RoundTeamStatus):
    name: Literal['Wind and Freedom'] = 'Wind and Freedom'
    version: Literal['4.1'] = '4.1'
    decrease_usage_when_trigger: bool = False


class Pankration(TeamStatusBase):
    name: Literal['Pankration!'] = 'Pankration!'
    desc: str = (
        'After a player announces the end of their Round first, the other '
        'player, who has yet to announce the end of their Round, draws 2 '
        'cards.'
    )
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost()
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def event_handler_DECLARE_ROUND_END(
        self, event: DeclareRoundEndEventArguments, match: Any
    ) -> List[DrawCardAction | RemoveObjectAction]:
        """
        When anyone declears round end, another draw 2 cards and remove self
        """
        return [
            DrawCardAction(
                player_idx = 1 - event.action.player_idx,
                number = 2,
                draw_if_filtered_not_enough = True
            ),
            RemoveObjectAction(
                object_position = self.position,
            )
        ]


EventCardTeamStatus = (
    FreshWindOfFreedom | ChangingShifts | IHaventLostYet | LeaveItToMe 
    | EnduringRock | WhereIstheUnseenRazor | SprawlingGreenery
    | ReviveOnCooldown | StoneAndContracts | AncientCourtyard
    | FatuiAmbusher | RhythmOfTheGreatDream | WhenTheCraneReturned
    | WindAndFreedom | Pankration
)



from typing import List, Literal

from ...consts import ELEMENT_TO_DAMAGE_TYPE

from ..charactor_base import ElementalBurstBase, ElementalNormalAttackBase
from ..hydro.mirror_maiden import MirrorMaiden as MM_3_7
from ..hydro.mirror_maiden import InfluxBlast as IB_3_7


class InfluxBlast(IB_3_7):
    damage: int = 3


class MirrorMaiden_3_3(MM_3_7):
    version: Literal['3.3']
    skills: List[
        ElementalNormalAttackBase | InfluxBlast | ElementalBurstBase
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Water Ball',
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            InfluxBlast(),
            ElementalBurstBase(
                name = 'Rippled Reflection',
                damage = 5,
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalBurstBase.get_cost(self.element, 3, 2)
            )
        ]



from enum import Enum
from typing import List, Any
from ..utils import BaseModel
from .consts import (
    DieColor, DamageType, DamageElementalType,
    ElementalReactionType, ElementType, ObjectPositionType, SkillType
)
from .struct import ObjectPosition, Cost


class ModifiableValueTypes(str, Enum):
    """
    Enum representing the type of a modifiable value. A modifiable value is a
    value that can be modified by other objects before being used.
    """
    INITIAL_DICE_COLOR = 'INITIAL_DICE_COLOR'
    REROLL = 'REROLL'
    COMBAT_ACTION = 'COMBAT_ACTION'
    COST = 'COST'

    # damage calculation is split into 4 parts: increase, multiply and 
    # decrease. damage will be first increased, then multiplied, 
    # then decreased.
    DAMAGE_ELEMENT_ENHANCE = 'DAMAGE_ELEMENT_ENHANCE'
    DAMAGE_INCREASE = 'DAMAGE_INCREASE'
    DAMAGE_MULTIPLY = 'DAMAGE_MULTIPLY'
    DAMAGE_DECREASE = 'DAMAGE_DECREASE'


class ModifiableValueBase(BaseModel):
    """
    Base class of modifiable values. It saves the value and can pass through
    value modifiers to update its value in-place.
    """
    type: ModifiableValueTypes
    original_value: Any = None
    position: ObjectPosition

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.original_value is None:
            self.original_value = self.copy(deep = True)


class InitialDiceColorValue(ModifiableValueBase):
    type: ModifiableValueTypes = ModifiableValueTypes.INITIAL_DICE_COLOR
    player_idx: int
    dice_colors: List[DieColor] = []


class RerollValue(ModifiableValueBase):
    type: ModifiableValueTypes = ModifiableValueTypes.REROLL
    value: int
    player_idx: int


class CostValue(ModifiableValueBase):
    type: ModifiableValueTypes = ModifiableValueTypes.COST
    target_position: ObjectPosition | None
    cost: Cost

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        self.cost = self.cost.copy(deep = True)
        self.original_value.cost = self.original_value.cost.copy(deep = True)
        self.original_value.cost.original_value = None
        self.cost.original_value = self.original_value.cost.copy(deep = True)


class CombatActionValue(ModifiableValueBase):
    type: ModifiableValueTypes = ModifiableValueTypes.COMBAT_ACTION
    action_label: int
    do_combat_action: bool


class DamageElementEnhanceValue(ModifiableValueBase):
    """
    It describes a detailed damage, i.e. charactor X will receive the damage.
    It contains all information that a damage is needed, except element
    reaction, as it is not decided yet.

    It is the initial stage of damage calculation.
    In this stage, buffs will modify the damage elemental type.
    """
    type: ModifiableValueTypes = ModifiableValueTypes.DAMAGE_ELEMENT_ENHANCE
    damage_type: DamageType
    target_position: ObjectPosition
    damage: int
    damage_elemental_type: DamageElementalType
    cost: Cost  # original cost of source

    damage_from_element_reaction: bool = False

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.damage_type == DamageType.HEAL:
            assert self.damage < 0, 'Heal should be negative'
            assert self.damage_elemental_type == DamageElementalType.HEAL, (
                'Heal should be heal type'
            )
        elif self.damage_type == DamageType.DAMAGE:
            assert self.damage >= 0, 'Damage should be non-negative'
        else:
            assert self.damage_type == DamageType.ELEMENT_APPLICATION
            assert self.damage == 0, 'Element application should be 0'

    def is_corresponding_charactor_use_damage_skill(
        self,
        object_position: ObjectPosition,
        match: Any,
        skill_type: SkillType | None
    ) -> bool:
        """
        Most objects modifying damage when current charactor
        makes damages by using skills. This function checks if the object
        with given object_position is the corresponding charactor using skill.

        If object position is on certain charactor, i.e. equipments or 
        charactor status, the charactor should match damage source. Otherwise,
        the active charactor should be the damage source.

        Args:
            object_position: The position of the object to check.
            match: The match object.
            skill_type: If not None, the skill type should match.
        """
        if self.damage_type != DamageType.DAMAGE:
            # not damage
            return False
        if self.position.area != ObjectPositionType.SKILL:
            # not skill
            return False
        if self.position.player_idx != object_position.player_idx:
            # not same player
            return False
        if (
            object_position.area == ObjectPositionType.CHARACTOR
            or object_position.area == ObjectPositionType.CHARACTOR_STATUS
        ):
            # object on charactor, get idx
            charactor_idx = object_position.charactor_idx
        else:
            # not on charactor, get active charactor idx
            charactor_idx = match.player_tables[
                object_position.player_idx].active_charactor_idx
        if self.position.charactor_idx != charactor_idx:
            # not same charactor
            return False
        if skill_type is not None:
            skill = match.get_object(self.position)
            if skill.skill_type != skill_type:
                # skill type not match
                return False
        return True

    def is_corresponding_charactor_receive_damage(
        self,
        object_position: ObjectPosition,
        match: Any,
        ignore_piercing: bool = True
    ) -> bool:
        """
        Most objects modifing damage only when current charactor
        receives damages. This function checks if the object with given
        object_position is the corresponding charactor receiving damage.

        If object position is on certain charactor, i.e. equipments or 
        charactor status, the charactor should match damage target. Otherwise,
        the active charactor should be the damage target.

        Args:
            object_position: The position of the object to check.
            match: The match object.
            ignore_piercing: If True, piercing damage will result in False.
        """
        if self.damage_type != DamageType.DAMAGE:
            # not damage
            return False
        if (
            ignore_piercing 
            and self.damage_elemental_type == DamageElementalType.PIERCING
        ):
            # piercing damage
            return False
        if (
            self.target_position.area != ObjectPositionType.CHARACTOR
        ):  # pragma: no cover
            # not on charactor
            return False
        if self.target_position.player_idx != object_position.player_idx:
            # not same player
            return False
        if (
            object_position.area == ObjectPositionType.CHARACTOR
            or object_position.area == ObjectPositionType.CHARACTOR_STATUS
        ):
            # object on charactor, get idx
            charactor_idx = object_position.charactor_idx
        else:
            # not on charactor, get active charactor idx
            charactor_idx = match.player_tables[
                object_position.player_idx].active_charactor_idx
        if self.target_position.charactor_idx != charactor_idx:
            # not same charactor
            return False
        return True


class DamageIncreaseValue(DamageElementEnhanceValue):
    """
    It inherits DamageElementEnhanceValue, and add two information about
    elemental reaction.

    In this stage, damage elemental type has been decided, and all add-type
    damage buffs will modify the damage, including elemental reaction.
    """
    type: ModifiableValueTypes = ModifiableValueTypes.DAMAGE_INCREASE

    element_reaction: ElementalReactionType
    reacted_elements: List[ElementType]

    @staticmethod
    def from_element_enhance_value(
        value: DamageElementEnhanceValue,
        element_reaction: ElementalReactionType,
        reacted_elements: List[ElementType],
    ) -> 'DamageIncreaseValue':
        return DamageIncreaseValue(
            position = value.position,
            damage_type = value.damage_type,
            target_position = value.target_position,
            damage = value.damage,
            damage_elemental_type = value.damage_elemental_type,
            cost = value.cost,
            damage_from_element_reaction = value.damage_from_element_reaction,
            element_reaction = element_reaction,
            reacted_elements = reacted_elements,
        )


class DamageMultiplyValue(DamageIncreaseValue):
    """
    In this stage, damage increase has been decided, and all multiply-type
    buffs will modify the damage, e.g. Mona elemental burst, Noelle elemental
    skill.
    """
    type: ModifiableValueTypes = ModifiableValueTypes.DAMAGE_MULTIPLY

    @staticmethod
    def from_increase_value(
        increase_value: DamageIncreaseValue
    ) -> 'DamageMultiplyValue':
        return DamageMultiplyValue(
            position = increase_value.position,
            damage_type = increase_value.damage_type,
            target_position = increase_value.target_position,
            damage = increase_value.damage,
            damage_elemental_type = increase_value.damage_elemental_type,
            damage_from_element_reaction = 
            increase_value.damage_from_element_reaction,
            cost = increase_value.cost,
            element_reaction = increase_value.element_reaction,
            reacted_elements = increase_value.reacted_elements,
        )


class DamageDecreaseValue(DamageIncreaseValue):
    """
    In this stage, damage multiply has been decided, and all decrease-type
    buffs will modify the damage, i.e. various types of shields.
    """
    type: ModifiableValueTypes = ModifiableValueTypes.DAMAGE_DECREASE

    @staticmethod
    def from_multiply_value(
        multiply_value: DamageMultiplyValue
    ) -> 'DamageDecreaseValue':
        return DamageDecreaseValue(
            position = multiply_value.position,
            damage_type = multiply_value.damage_type,
            target_position = multiply_value.target_position,
            damage = multiply_value.damage,
            damage_elemental_type = multiply_value.damage_elemental_type,
            damage_from_element_reaction = 
            multiply_value.damage_from_element_reaction,
            cost = multiply_value.cost,
            element_reaction = multiply_value.element_reaction,
            reacted_elements = multiply_value.reacted_elements,
        )

    def apply_shield(
        self, shield_usage: int, min_damage_to_trigger: int, 
        max_in_one_time: int, decrease_usage_by_damage: bool
    ) -> int:
        """
        Apply shield to damage, and return the shield usage after applying.

        Args:
            shield_usage: Current shield usage.
            decrease_usage_by_damage: If True, decrease usage by damage,
                and the following two parameters are ignored.
            min_damage_to_trigger: Minimum damage to trigger shield.
            max_in_one_time: Maximum damage to decrease in one time.
        """
        if shield_usage <= 0:
            # no usage, return
            return shield_usage
        if decrease_usage_by_damage:
            # decrease usage by damage
            decrease = min(shield_usage, self.damage)
            self.damage -= decrease
            shield_usage -= decrease
        else:
            if self.damage < min_damage_to_trigger:
                # damage too small to trigger
                return shield_usage
            decrease = min(max_in_one_time, self.damage)
            self.damage -= decrease
            shield_usage -= 1
        return shield_usage


DamageValue = DamageElementEnhanceValue
FinalDamageValue = DamageDecreaseValue  # alias



from typing import Any, List, Literal

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Charactor status. DO NOT define here, define in server/status/characor_status
# Here is just example.


# Round status, will last for several rounds and disappear


# Skills


class SacredRiteHeronsSanctum(ElementalSkillBase):
    name: Literal[
        "Sacred Rite: Heron's Sanctum"] = "Sacred Rite: Heron's Sanctum"
    desc: str = (
        'Attaches a Heron Shield to this character and prepares Heron Strike'
    )
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        prepare
        """
        return [
            self.charge_self(1),
            self.create_charactor_status('Heron Shield'),
        ]


class HeronStrike(ElementalSkillBase):
    name: Literal['Heron Strike'] = 'Heron Strike'
    desc: str = '''(Prepare for 1 turn) Deals 3 Hydro DMG.'''
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost()

    def is_valid(self, match: Any) -> bool:
        """
        alway invalid to directly use.
        """
        return False

    def get_actions(self, match: Any) -> List[Actions]:
        """
        no charge, only attack
        """
        return [
            self.attack_opposite_active(match, self.damage, self.damage_type)
        ]


class SacredRiteWagtailsTide(ElementalBurstBase):
    name: Literal[
        "Sacred Rite: Wagtail's Tide"] = "Sacred Rite: Wagtail's Tide"
    desc: str = '''Deals 2 Hydro DMG and create Prayer of the Crimson Crown.'''
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_team_status('Prayer of the Crimson Crown'),
        ]


# Talents


class TheOverflow(SkillTalent):
    name: Literal['The Overflow']
    desc: str = (
        'Combat Action: When your active character is Candace, equip this '
        'card. After Candace equips this card, immediately use Sacred Rite: '
        "Wagtail's Tide once. When this card is equipped by Candace, her "
        'Prayer of the Crimson Crown has the following extra effect: After '
        'your character uses a Normal Attack: Deals 1 Hydro DMG. (Once per '
        'Round)'
    )
    version: Literal['3.8'] = '3.8'
    charactor_name: Literal['Candace'] = 'Candace'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 4,
        charge = 2
    )
    skill: SacredRiteWagtailsTide = SacredRiteWagtailsTide()


# charactor base


class Candace(CharactorBase):
    name: Literal['Candace']  # Do not set default value for charactor name
    version: Literal['3.8'] = '3.8'
    desc: str = '''"Golden Vow" Candace'''
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | SacredRiteHeronsSanctum 
        | SacredRiteWagtailsTide | HeronStrike
    ] = []
    faction: List[FactionType] = [
        FactionType.SUMERU
    ]
    weapon_type: WeaponType = WeaponType.POLEARM
    talent: TheOverflow | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Gleaming Spear - Guardian Stance',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SacredRiteHeronsSanctum(),
            SacredRiteWagtailsTide(),
            HeronStrike(),
        ]



from src.lpsim.agents.interaction_agent import InteractionAgent
from src.lpsim.agents.nothing_agent import NothingAgent
from src.lpsim.server.match import Match, MatchState
from src.lpsim.server.consts import DamageElementalType
from src.lpsim.server.deck import Deck
from tests.utils_for_test import (
    check_hp, get_test_id_from_command, make_respond, get_random_state,
    set_16_omni
)


def test_fischl_mona_nahida():
    """
    3335 + E + talent
    """
    agent_0 = NothingAgent(player_idx = 0)
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 0",
            "card 1 0 0 1",
            "skill 1 0 1 2",
            "sw_char 1 1",
            "skill 1 omni omni hydro",
            "end",
            "TEST 1 hp 4 8 8",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "skill 2 omni omni dendro omni omni",
            "TEST 2 hp 0 7 7 and 1 seed for alive",
            "end",
            "TEST 3 hp 0 1 5",
            "skill 1 0 1 2",
            "card 0 0 0 1 2"
        ],
        only_use_command = True
    )
    match = Match(random_state = get_random_state())
    deck = Deck.from_str(
        """
        charactor:Fischl
        charactor:Mona
        charactor:Nahida
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    set_16_omni(match)
    assert match.start()
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    check_hp(match, [[4, 8, 8], [10, 10, 10]])
                elif test_id == 2:
                    check_hp(match, [[0, 7, 7], [10, 10, 10]])
                    for charactor in match.player_tables[0].charactors:
                        if charactor.is_alive:
                            assert len(charactor.status) == 1
                            assert charactor.status[
                                0].name == "Seed of Skandha"
                        else:
                            assert len(charactor.status) == 0
                elif test_id == 3:
                    check_hp(match, [[0, 1, 5], [10, 10, 10]])
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError('No need respond.')
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[0, 0, 1], [10, 10, 10]])
    assert len(match.player_tables[1].team_status) == 1
    assert match.player_tables[1].team_status[0].name == 'Shrine of Maya'
    assert match.player_tables[1].team_status[0].usage == 3

    assert match.state != MatchState.ERROR


def test_fischl_mona_nahida_no_talent():
    """
    3335 + EQ
    """
    agent_0 = NothingAgent(player_idx = 0)
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 0",
            "card 1 0 0 1",
            "skill 1 0 1 2",
            "sw_char 1 1",
            "skill 1 omni omni hydro",
            "end",
            "TEST 1 hp 4 8 8",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "skill 2 omni omni dendro omni omni",
            "TEST 2 hp 0 7 7 and 1 seed for alive",
            "end",
            "TEST 3 hp 0 1 5",
            "skill 1 0 1 2",
            "skill 3 0 1 2"
        ],
        only_use_command = True
    )
    match = Match(random_state = get_random_state())
    deck = Deck.from_str(
        """
        charactor:Fischl
        charactor:Mona
        charactor:Nahida
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    set_16_omni(match)
    assert match.start()
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    check_hp(match, [[4, 8, 8], [10, 10, 10]])
                elif test_id == 2:
                    check_hp(match, [[0, 7, 7], [10, 10, 10]])
                    for charactor in match.player_tables[0].charactors:
                        if charactor.is_alive:
                            assert len(charactor.status) == 1
                            assert charactor.status[
                                0].name == "Seed of Skandha"
                        else:
                            assert len(charactor.status) == 0
                elif test_id == 3:
                    check_hp(match, [[0, 1, 5], [10, 10, 10]])
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError('No need respond.')
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[0, 0, 1], [10, 10, 10]])
    # maks small tests on getting id
    assert match.player_tables[0].previous_charactor_idx() is None
    assert match.player_tables[0].next_charactor_idx() is None
    assert match.player_tables[1].previous_charactor_idx() == 1
    assert match.player_tables[1].next_charactor_idx() == 0
    for i in range(3):
        assert match.player_tables[1].previous_charactor_idx(i) == (i + 2) % 3
        assert match.player_tables[1].next_charactor_idx(i) == (i + 1) % 3
    assert len(match.player_tables[1].team_status) == 1
    assert match.player_tables[1].team_status[0].name == 'Shrine of Maya'
    assert match.player_tables[1].team_status[0].usage == 2

    assert match.state != MatchState.ERROR

    assert match.need_respond(1)
    agent_1.commands = ['skill 1 0 1 2']
    make_respond(agent_1, match)
    assert match.state == MatchState.ENDED
    assert not match.step()


def test_nahida_talents():
    """
    first: with no pyro hydro electro ally
    """
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 0",
            "sw_char 1 0",
            "sw_char 0 0",
            "end",
            "sw_char 2 0",
            "sw_char 0 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "end",
            "end"
        ],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 2",
            "skill 1 0 1 2",
            "TEST 1 only p0c1 2seed",
            "skill 1 0 1 2",
            "TEST 2 p0c0 p0c1 2seed",
            "skill 1 0 1 2",
            "TEST 3 all 2seed hp 6 8 10",
            "end",
            "card 5 0 0 1 2",
            "TEST 4 maya 2 seed all 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "TEST 5 hp 5 4 5 mid no element",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "TEST 6 hp 5 2 5 all dendro 2seed",
            "sw_char 1 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "end",
            "TEST 7 hp 4 1 1, ele DDN, all 1seed, burining flame, 1 maya",
            "sw_char 0 0"
        ],
        only_use_command = True
    )
    match = Match(version = '0.0.1', random_state = get_random_state())
    deck1 = Deck.from_str('''
        default_version:4.0
        charactor:ElectroMobMage
        charactor:PyroMobMage
        charactor:HydroMobMage
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
    ''')
    deck2 = Deck.from_str(
        """
        charactor:DendroMobMage
        charactor:DendroMobMage
        charactor:Nahida
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
        """
    )
    match.set_deck([deck1, deck2])
    match.config.max_same_card_number = 30
    match.config.check_deck_restriction = False
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()
    match.player_tables[1].charactors[0].skills[
        0].damage_type = DamageElementalType.PYRO  # can do pyro damage
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    charactors = match.player_tables[0].charactors
                    assert len(charactors[0].status) == 0
                    assert len(charactors[1].status) == 1
                    assert charactors[1].status[0].name == "Seed of Skandha"
                    assert charactors[1].status[0].usage == 2
                    assert len(charactors[2].status) == 0
                elif test_id == 2:
                    charactors = match.player_tables[0].charactors
                    assert len(charactors[0].status) == 1
                    assert charactors[0].status[0].name == "Seed of Skandha"
                    assert charactors[0].status[0].usage == 2
                    assert len(charactors[1].status) == 1
                    assert charactors[1].status[0].name == "Seed of Skandha"
                    assert charactors[1].status[0].usage == 2
                    assert len(charactors[2].status) == 0
                elif test_id == 3:
                    charactors = match.player_tables[0].charactors
                    for charactor in charactors:
                        assert len(charactor.status) == 1
                        assert charactor.status[0].name == "Seed of Skandha"
                        assert charactor.status[0].usage == 2
                    for charactor in charactors[:2]:
                        assert charactor.element_application == ['DENDRO']
                    assert charactors[2].element_application == []
                    check_hp(match, [[6, 8, 10], [10, 10, 10]])
                elif test_id == 4:
                    charactors = match.player_tables[0].charactors
                    for charactor in charactors:
                        assert len(charactor.status) == 1
                        assert charactor.status[0].name == "Seed of Skandha"
                        assert charactor.status[0].usage == 2
                    for charactor in charactors:
                        assert charactor.element_application == ['DENDRO']
                    assert len(match.player_tables[1].team_status) == 1
                    assert match.player_tables[1].team_status[0].name == (
                        'Shrine of Maya'
                    )
                    assert match.player_tables[1].team_status[0].usage == 2
                elif test_id == 5:
                    charactors = match.player_tables[0].charactors
                    for cid, charactor in enumerate(charactors):
                        assert len(charactor.status) == 1
                        assert charactor.status[0].name == "Seed of Skandha"
                        assert charactor.status[0].usage == 1
                        if cid == 1:
                            assert charactor.element_application == []
                        else:
                            assert charactor.element_application == ['DENDRO']
                    check_hp(match, [[5, 4, 5], [10, 10, 10]])
                elif test_id == 6:
                    # TEST 6 hp 5 2 5 all dendro 2seed
                    charactors = match.player_tables[0].charactors
                    for charactor in charactors:
                        assert len(charactor.status) == 1
                        assert charactor.status[0].name == "Seed of Skandha"
                        assert charactor.status[0].usage == 2
                        assert charactor.element_application == ['DENDRO']
                    check_hp(match, [[5, 2, 5], [10, 10, 10]])
                elif test_id == 7:
                    # hp 4 1 1, ele DDN, all 1seed, burining flame, 1 maya
                    charactors = match.player_tables[0].charactors
                    for cid, charactor in enumerate(charactors):
                        assert len(charactor.status) == 1
                        assert charactor.status[0].name == "Seed of Skandha"
                        assert charactor.status[0].usage == 1
                        if cid == 2:
                            assert charactor.element_application == []
                        else:
                            assert charactor.element_application == ['DENDRO']
                    check_hp(match, [[4, 1, 1], [10, 7, 10]])
                    assert len(match.player_tables[1].team_status) == 1
                    assert match.player_tables[1].team_status[0].name == (
                        'Shrine of Maya'
                    )
                    assert match.player_tables[1].team_status[0].usage == 1
                    assert len(match.player_tables[1].summons) == 1
                    assert match.player_tables[1].summons[0].name == (
                        'Burning Flame'
                    )
                    assert match.player_tables[1].summons[0].usage == 1
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError('No need respond.')
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[4, 1, 1], [10, 7, 10]])

    assert match.state != MatchState.ERROR


def test_nahida_talents_2():
    """
    second: with pyro electro (hydro tested above)
    """
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 1",
            "skill 1 0 1 2",
            "sw_char 0 0",
            "end",
            "sw_char 2 0",
            "sw_char 0 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "sw_char 2 0",
            "end",
            "end"
        ],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 2",
            "skill 1 0 1 2",
            "TEST 1 only p0c1 2seed",
            "skill 1 0 1 2",
            "TEST 2 p0c0 p0c1 2seed",
            "skill 1 0 1 2",
            "TEST 3 all 2seed hp 6 8 10",
            "end",
            "card 5 0 0 1 2",
            "TEST 4 maya 2 seed all 3, p1c2 seed 2",
            "skill 1 0 1 2",
            "TEST 4 maya 2 seed all 3, p1c2 seed 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "TEST 5 hp 3 4 5 all dendro",
            "sw_char 2 0",
            "end",
            "TEST 7 hp 4 3 1, all dendro, all 1seed, burining flame, 1 maya",
            "sw_char 0 0"
        ],
        only_use_command = True
    )
    match = Match(version = '0.0.1', random_state = get_random_state())
    deck1 = Deck.from_str('''
        default_version:4.0
        charactor:ElectroMobMage
        charactor:Nahida
        charactor:HydroMobMage
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
    ''')
    deck2 = Deck.from_str(
        """
        charactor:PyroMobMage
        charactor:ElectroMobMage
        charactor:Nahida
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
        """
    )
    match.set_deck([deck1, deck2])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    charactors = match.player_tables[0].charactors
                    assert len(charactors[0].status) == 0
                    assert len(charactors[1].status) == 1
                    assert charactors[1].status[0].name == "Seed of Skandha"
                    assert charactors[1].status[0].usage == 2
                    assert len(charactors[2].status) == 0
                elif test_id == 2:
                    charactors = match.player_tables[0].charactors
                    assert len(charactors[0].status) == 1
                    assert charactors[0].status[0].name == "Seed of Skandha"
                    assert charactors[0].status[0].usage == 2
                    assert len(charactors[1].status) == 1
                    assert charactors[1].status[0].name == "Seed of Skandha"
                    assert charactors[1].status[0].usage == 2
                    assert len(charactors[2].status) == 0
                elif test_id == 3:
                    charactors = match.player_tables[0].charactors
                    for charactor in charactors:
                        assert len(charactor.status) == 1
                        assert charactor.status[0].name == "Seed of Skandha"
                        assert charactor.status[0].usage == 2
                    for charactor in charactors[:2]:
                        assert charactor.element_application == ['DENDRO']
                    assert charactors[2].element_application == []
                    check_hp(match, [[6, 8, 10], [10, 10, 8]])
                elif test_id == 4:
                    charactors = match.player_tables[0].charactors
                    for charactor in charactors:
                        assert len(charactor.status) == 1
                        assert charactor.status[0].name == "Seed of Skandha"
                        assert charactor.status[0].usage == 3
                    for charactor in charactors:
                        assert charactor.element_application == ['DENDRO']
                    assert len(match.player_tables[1].team_status) == 1
                    assert match.player_tables[1].team_status[0].name == (
                        'Shrine of Maya'
                    )
                    assert match.player_tables[1].team_status[0].usage == 2
                    assert match.player_tables[1].charactors[2].status[
                        0].name == 'Seed of Skandha'
                    assert match.player_tables[1].charactors[2].status[
                        0].usage == 2
                elif test_id == 5:
                    charactors = match.player_tables[0].charactors
                    for cid, charactor in enumerate(charactors):
                        assert len(charactor.status) == 1
                        assert charactor.status[0].name == "Seed of Skandha"
                        assert charactor.status[0].usage == 2
                        assert charactor.element_application == ['DENDRO']
                    check_hp(match, [[3, 4, 5], [10, 10, 8]])
                elif test_id == 7:
                    # hp 4 1 1, ele DDN, all 1seed, burining flame, 1 maya
                    charactors = match.player_tables[0].charactors
                    for cid, charactor in enumerate(charactors):
                        assert len(charactor.status) == 1
                        assert charactor.status[0].name == "Seed of Skandha"
                        assert charactor.status[0].usage == 1
                        assert charactor.element_application == ['DENDRO']
                    check_hp(match, [[2, 3, 1], [10, 10, 8]])
                    assert len(match.player_tables[1].team_status) == 1
                    assert match.player_tables[1].team_status[0].name == (
                        'Shrine of Maya'
                    )
                    assert match.player_tables[1].team_status[0].usage == 1
                    assert len(match.player_tables[1].summons) == 1
                    assert match.player_tables[1].summons[0].name == (
                        'Burning Flame'
                    )
                    assert match.player_tables[1].summons[0].usage == 1
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError('No need respond.')
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[2, 3, 1], [10, 10, 8]])

    assert match.state != MatchState.ERROR


def test_nahida_apply_seed_to_defeated():
    """
    try to apply seed to defeated charactor.
    """
    agent_0 = NothingAgent(player_idx = 0)
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 2",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "skill 3 0 1 2",
            "skill 1 0 1 2",
            "end",
            "skill 1 0 1 2",
            "skill 2 0 1 2 3 4",
            "end",
        ],
        only_use_command = True
    )
    match = Match(random_state = get_random_state())
    deck = Deck.from_str(
        """
        charactor:Fischl
        charactor:Mona
        charactor:Nahida
        Rana*10
        Wine-Stained Tricorne*10
        The Seed of Stored Knowledge*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    set_16_omni(match)
    assert match.start()
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError('No need respond.')
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[0, 5, 10], [10, 10, 10]])
    assert len(match.player_tables[1].team_status) == 0
    for charactor in match.player_tables[0].charactors:
        if charactor.is_alive:
            assert len(charactor.status) == 1
            assert charactor.status[0].name == "Seed of Skandha"
            assert charactor.status[0].usage == 2
        else:
            assert len(charactor.status) == 0

    assert match.state != MatchState.ERROR


def test_maya_not_first_status():
    """
    AA + electro A + talent
    """
    agent_0 = NothingAgent(player_idx = 0)
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 2",
            "skill 0 0 1 2",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "end",
            "card 0 0 0 1 2",
            "TEST 1 maya 3",
            "end",
        ],
        only_use_command = True
    )
    match = Match(random_state = get_random_state())
    deck = Deck.from_str(
        """
        charactor:ElectroMobMage
        charactor:Mona
        charactor:Nahida
        The Seed of Stored Knowledge*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    set_16_omni(match)
    assert match.start()
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    status = match.player_tables[1].team_status
                    found = False
                    for s in status:
                        if s.name == 'Shrine of Maya':
                            assert s.usage == 3
                            found = True
                    assert found
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError('No need respond.')
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[1, 10, 10], [10, 10, 10]])
    assert len(match.player_tables[1].team_status) == 2
    assert match.player_tables[1].team_status[1].name == 'Shrine of Maya'
    assert match.player_tables[1].team_status[1].usage == 2
    assert match.player_tables[1].team_status[0].name == 'Catalyzing Field'
    assert match.player_tables[1].team_status[0].usage == 1

    assert match.state != MatchState.ERROR


def test_seed_not_first_status():
    """
    AA + electro A + talent
    """
    agent_0 = NothingAgent(player_idx = 0)
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 1",
            "skill 0 0 1 2",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "sw_char 1 0",
            "skill 0 0 1 2",
            "sw_char 2 0",
        ],
        only_use_command = True
    )
    match = Match(random_state = get_random_state())
    deck = Deck.from_str(
        """
        charactor:CryoMobMage
        charactor:Mona
        charactor:Nahida
        The Seed of Stored Knowledge*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    set_16_omni(match)
    match.config.initial_dice_number = 32
    match.config.max_dice_number = 32
    assert match.start()
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError('No need respond.')
        if len(agent_1.commands) == 0:
            break

    assert len(agent_1.commands) == 0
    assert match.round_number == 1
    check_hp(match, [[0, 9, 9], [10, 10, 10]])
    charactors = match.player_tables[0].charactors
    assert len(charactors[0].status) == 0
    assert len(charactors[1].status) == 1
    assert charactors[1].status[0].name == 'Seed of Skandha'
    assert charactors[1].status[0].usage == 1
    assert len(charactors[2].status) == 1
    assert charactors[2].status[0].name == 'Seed of Skandha'
    assert charactors[2].status[0].usage == 1

    assert match.state != MatchState.ERROR


def test_talent_enemy_has_other_charactor_status():
    """
    when enemy has other charactor status, equipping talent should only
    add usage of seed.
    in round 3, player 0 will do elemental reaction but will not increase
    damage.
    """
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 0",
            "end",
            "end",
            "skill 0 0 1 2",
            "sw_char 1 0",
            "skill 0 0 1 2",
        ],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = [
            "sw_card",
            "choose 1",
            "skill 0 0 1 2",
            "sw_char 3 0",
            "skill 1 0 1 2",
            "skill 1 0 1 2",
            "sw_char 2 0",
            "end",
            "skill 0 0 1 2",
            "sw_char 3 0",
            "card 0 0 0 1 2",
            "TEST 1 maya 3 seed all 2 frozen 1",
            "skill 1 0 1 2",
            "end",
            "end",
        ],
        only_use_command = True
    )
    match = Match(random_state = get_random_state())
    deck = Deck.from_str(
        """
        charactor:ElectroMobMage
        charactor:CryoMobMage
        charactor:Mona
        charactor:Nahida
        The Seed of Stored Knowledge*30
        """
    )
    # add HP to avoid defeated.
    deck.charactors[0].hp = 90
    deck.charactors[0].max_hp = 90
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.charactor_number = 4
    match.config.random_first_player = False
    set_16_omni(match)
    # match.enable_history = True
    assert match.start()
    match.step()

    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            while True:
                test_id = get_test_id_from_command(agent_1)
                if test_id == 1:
                    status = match.player_tables[1].team_status
                    found = False
                    for s in status:
                        if s.name == 'Shrine of Maya':  # pragma: no branch
                            assert s.usage == 3
                            found = True
                    assert found
                    charactors = match.player_tables[0].charactors
                    for charactor in charactors:
                        assert charactor.status[0].name == "Seed of Skandha"
                        assert charactor.status[0].usage == 2
                    assert len(charactors[0].status) == 2
                    assert charactors[0].status[1].name == "Frozen"
                    assert charactors[0].status[1].usage == 1
                else:
                    break
            make_respond(agent_1, match)
        else:
            raise AssertionError('No need respond.')
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    assert len(agent_1.commands) == 0 and len(agent_0.commands) == 0
    assert match.round_number == 3
    check_hp(match, [[76, 9, 9, 9], [89, 9, 9, 7]])
    assert len(match.player_tables[1].team_status) == 1
    assert match.player_tables[1].team_status[0].name == 'Shrine of Maya'
    assert match.player_tables[1].team_status[0].usage == 2

    assert match.state != MatchState.ERROR


def test_seed_with_element_application():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "sw_char 2 0",
            "sw_char 1 0",
            "skill 2 0 1 2 3 4",
            "TEST 2 p0 team usage 1",
            "TEST 3 p1 charactor usage 2",
            "TEST 1 10 6 10 7 10 10",
            "end"
        ],
        [
            "sw_card",
            "choose 1",
            "sw_char 0 0",
            "skill 0 0 1 2",
            "skill 1 0 1 2"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.0
        charactor:Xingqiu@3.3
        charactor:Nahida
        charactor:Fischl
        The Scent Remained*30
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(' ')[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                assert len(match.player_tables[0].team_status) == 1
                assert match.player_tables[0].team_status[0].usage == 1
            elif test_id == 3:
                for c in match.player_tables[1].charactors:
                    assert len(c.status) == 1
                    assert c.status[0].usage == 2
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == '__main__':
    # test_fischl_mona_nahida()
    # test_seed_not_first_status()
    # test_talent_enemy_has_other_charactor_status()
    # test_nahida_talents()
    # test_seed_with_element_application()
    test_fischl_mona_nahida_no_talent()



"""
Team status generated by system.
"""

from typing import Any, Literal
from .base import ShieldTeamStatus, UsageTeamStatus
from ...modifiable_values import DamageIncreaseValue
from ...consts import DamageElementalType, DamageType, IconType


class CatalyzingField(UsageTeamStatus):
    """
    Catalyzing field.
    """
    name: Literal['Catalyzing Field'] = 'Catalyzing Field'
    desc: str = (
        'When you deal Electro DMG or Pyro DMG to an opposing active '
        'charactor, DMG dealt +1.'
    )
    version: Literal['3.4'] = '3.4'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def value_modifier_DAMAGE_INCREASE(
            self, value: DamageIncreaseValue, match: Any,
            mode: Literal['TEST', 'REAL']) -> DamageIncreaseValue:
        """
        Increase damage for dendro or electro damages, and decrease usage.
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, not modify
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
        ):
            # source not self, not activate
            return value
        if not self.position.check_position_valid(
            value.target_position, match, 
            player_idx_same = False, target_is_active_charactor = True,
        ):
            # target not enemy, or target not active charactor, not activate
            return value
        if value.damage_elemental_type in [
            DamageElementalType.DENDRO,
            DamageElementalType.ELECTRO,
        ] and self.usage > 0:
            value.damage += 1
            assert mode == 'REAL'
            self.usage -= 1
        return value


class DendroCore(UsageTeamStatus):
    """
    Dendro core.
    """
    name: Literal['Dendro Core'] = 'Dendro Core'
    desc: str = (
        'When you deal Pyro DMG or Electro DMG to an opposing active '
        'charactor, DMG dealt +2.'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def value_modifier_DAMAGE_INCREASE(
            self, value: DamageIncreaseValue, match: Any,
            mode: Literal['TEST', 'REAL']) -> DamageIncreaseValue:
        """
        Increase damage for electro or pyro damages by 2, and decrease usage.
        """
        if value.damage_type != DamageType.DAMAGE:
            # not damage, not modify
            return value
        if not self.position.check_position_valid(
            value.position, match, player_idx_same = True,
        ):
            # source not self, not activate
            return value
        if not self.position.check_position_valid(
            value.target_position, match, 
            player_idx_same = False, target_is_active_charactor = True,
        ):
            # target not enemy, or target not active charactor, not activate
            return value
        if value.damage_elemental_type in [
            DamageElementalType.ELECTRO,
            DamageElementalType.PYRO,
        ] and self.usage > 0:
            value.damage += 2
            assert mode == 'REAL'
            self.usage -= 1
        return value


class Crystallize(ShieldTeamStatus):
    """
    Crystallize.
    """
    name: Literal['Crystallize'] = 'Crystallize'
    desc: str = (
        'Grants 1 Shield point to your active charactor. '
        '(Can stack. Max 2 Points.)'
    )
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 2


SystemTeamStatus = CatalyzingField | DendroCore | Crystallize



from .others import OtherEventCards
from .arcane_legend import ArcaneLegendCards
from .resonance import ElementResonanceCards, NationResonanceCards
from .foods import FoodCards
from .old_version import OldVersionEventCards


EventCards = (
    ArcaneLegendCards | ElementResonanceCards | NationResonanceCards 
    | OtherEventCards | FoodCards
    # finally old version cards
    | OldVersionEventCards
)



from src.lpsim.agents import InteractionAgent
from src.lpsim import Deck, Match, MatchState
from tests.utils_for_test import (
    check_hp, check_usage, get_random_state, get_test_id_from_command, 
    make_respond, set_16_omni
)


def test_dehya():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 0 15 14 13",
            "TEST 1 10 10 10 10 10 9",
            "skill 0 12 11 10",
            "TEST 1 9 10 10 10 10 8",
            "sw_char 1 9",
            "sw_char 0 8",
            "skill 0 7 6 5",
            "skill 0 4 3 2",
            "TEST 1 9 10 10 10 9 7",
            "end",
            "skill 2 15 14 13",
            "sw_char 1 12",
            "skill 0 11 10 9",
            "sw_char 0 8",
            "end",
            "TEST 1 7 10 10 10 2 7",
            "skill 0 15 14 13",
            "sw_char 1 12",
            "TEST 1 7 9 10 9 2 6",
            "TEST 2 p1 summon 3",
            "sw_char 2 11",
            "sw_char 1 10",
            "skill 1 9 8 7",
            "TEST 1 7 4 10 9 2 5",
            "sw_char 2 6",
            "skill 1 5 4 3",
            "sw_char 0 2",
            "TEST 1 6 4 9 9 2 5",
            "end",
            "sw_char 1 15",
            "sw_char 0 14",
            "skill 0 13 12 11",
            "card 8 0",
            "skill 0 10 9 8",
            "skill 2 7 6 5",
            "TEST 1 6 4 9 2 2 1",
            "end",
            "sw_char 2 15",
            "skill 0 14 13 12",
            "sw_char 1 11",
            "skill 0 10 9 8",
            "end",
            "card 8 1",
            "skill 0 15 14 13",
            "sw_char 2 12",
            "card 3 0 11 10 9 8",
            "TEST 1 6 2 6 2 1 0",
            "sw_char 0 7",
            "card 2 0",
            "end",
            "sw_char 2 15",
            "card 7 2",
            "sw_char 0 14",
            "end",
            "sw_char 2 15",
            "sw_char 0 14",
            "TEST 1 7 2 7 2 1 0",
            "end"
        ],
        [
            "sw_card",
            "choose 2",
            "skill 1 15 14 13",
            "TEST 1 10 10 10 10 10 8",
            "skill 1 12 11 10",
            "sw_char 1 9",
            "TEST 1 9 10 10 10 10 7",
            "end",
            "sw_char 0 15",
            "sw_char 1 14",
            "TEST 1 8 10 10 10 3 6",
            "end",
            "card 6 1",
            "sw_char 0 15",
            "sw_char 1 14",
            "TEST 1 7 10 10 9 2 6",
            "sw_char 2 13",
            "skill 1 12 11 10",
            "skill 2 9 8 7 6",
            "sw_char 0 5",
            "TEST 1 7 4 10 9 2 5",
            "sw_char 2 4",
            "skill 0 3 2 1",
            "end",
            "card 6 2",
            "sw_char 0 15",
            "card 2 1",
            "card 3 0",
            "end",
            "sw_char 2 15",
            "skill 1 14 13 12",
            "choose 0",
            "card 0 0",
            "card 3 0",
            "sw_char 1 11",
            "skill 0 10 9 8",
            "TEST 1 6 2 6 2 2 0",
            "sw_char 0 7",
            "end",
            "card 7 0",
            "sw_char 1 15",
            "TEST 1 6 2 6 2 1 0",
            "sw_char 0 14",
            "end",
            "TEST 1 7 2 8 1 1 0",
            "card 9 0",
            "sw_char 1 15",
            "skill 0 14 13 12",
            "sw_char 0 11",
            "end",
            "TEST 1 7 2 8 1 1 0",
            "sw_char 1 15",
            "skill 0 14 13 12",
            "sw_char 0 11",
            "card 9 0",
            "end",
            "TEST 1 7 2 7 1 1 0",
            "end"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.1
        charactor:Wanderer
        charactor:Mona
        charactor:Dehya
        # Gales of Reverie*10
        # Beneficent*10
        Stalwart and True*10
        Sweet Madame*10
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(' ')
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx = int(cmd[2][1])
                summon = match.player_tables[pidx].summons
                check_usage(summon, cmd[4:])
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_dehya_2():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 15 14 13",
            "sw_char 2 12",
            "skill 1 11 10 9",
            "skill 1 8 7 6",
            "sw_char 0 5",
            "TEST 2 p1 team usage",
            "skill 1 4 3 2",
            "end",
            "skill 1 15 14 13",
            "end",
            "skill 1 15 14 13",
            "end",
            "sw_char 2 15",
            "skill 1 14 13 12",
            "sw_char 1 11",
            "sw_char 0 10",
            "skill 1 9 8 7",
            "sw_char 2 6",
            "skill 0 5 4 3",
            "sw_char 0 2",
            "TEST 1 10 10 9 10 3 4",
            "TEST 2 p0 team usage 2",
            "sw_char 2 1",
            "TEST 1 10 10 8 10 3 4",
            "end"
        ],
        [
            "sw_card",
            "choose 2",
            "skill 1 15 14 13",
            "sw_char 1 12",
            "sw_char 0 11",
            "sw_char 1 10",
            "TEST 1 10 10 10 10 10 8",
            "sw_char 2 9",
            "skill 1 8 7 6",
            "sw_char 1 5",
            "end",
            "end",
            "end",
            "TEST 2 p0 team usage 1 2",
            "sw_char 2 15",
            "sw_char 1 14",
            "sw_char 2 13",
            "sw_char 1 12",
            "skill 0 11 10 9",
            "sw_char 0 8",
            "sw_char 2 7",
            "skill 1 6 5 4",
            "skill 2 3 2 1 0"
        ]
    ]
    agent_0 = InteractionAgent(
        player_idx = 0,
        verbose_level = 0,
        commands = cmd_records[0],
        only_use_command = True
    )
    agent_1 = InteractionAgent(
        player_idx = 1,
        verbose_level = 0,
        commands = cmd_records[1],
        only_use_command = True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state = get_random_state())
    # deck information
    deck = Deck.from_str(
        '''
        default_version:4.1
        charactor:Noelle
        charactor:Mona
        charactor:Dehya
        # Gales of Reverie*10
        # Beneficent*10
        Stalwart and True*10
        Sweet Madame*10
        '''
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.charactor_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(' ')
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx = int(cmd[2][1])
                status = match.player_tables[pidx].team_status
                check_usage(status, cmd[5:])
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == '__main__':
    # test_dehya()
    test_dehya_2()



from .emblem_of_severed_fate import EmblemOfSeveredFateArtifacts
from .exile import ExileArtifacts
from .vermillion_shimenawa import VermillionShimenawas
from .heal_by_skill import HealBySkillArtifacts
from .element_artifacts import ElementArtifacts
from .millelith import MillelithArtifacts
from .gamblers import Gamblers
from .instructors_cap import InstructorsCap
from .ocean_hued import OceanHuedArtifacts

from .old_version import OldVersionArtifacts

Artifacts = (
    HealBySkillArtifacts | Gamblers | InstructorsCap | ExileArtifacts
    | EmblemOfSeveredFateArtifacts | MillelithArtifacts | VermillionShimenawas 
    | OceanHuedArtifacts | ElementArtifacts
    # finally old versions
    | OldVersionArtifacts
)



from typing import Any, List, Literal

from ...summon.base import AttackerSummonBase

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    AOESkillBase, ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class ClusterbloomArrow(AttackerSummonBase):
    name: Literal['Clusterbloom Arrow'] = 'Clusterbloom Arrow'
    desc: str = '''End Phase: Deal 1 Dendro DMG. (Can stack. Max 2 stacks.)'''
    version: Literal['3.6'] = '3.6'
    usage: int = 1
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.DENDRO
    damage: int = 1
    renew_type: Literal['ADD'] = 'ADD'


# Skills


class VijnanaPhalaMine(ElementalSkillBase):
    name: Literal['Vijnana-Phala Mine'] = 'Vijnana-Phala Mine'
    desc: str = (
        'Deals 2 Dendro DMG. This character gains Vijnana Suffusion.'
    )
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_charactor_status('Vijnana Suffusion'),
        ]


class FashionersTanglevineShaft(ElementalBurstBase, AOESkillBase):
    name: Literal[
        "Fashioner's Tanglevine Shaft"] = "Fashioner's Tanglevine Shaft"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    back_damage: int = 1
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3,
        charge = 2
    )


# Talents


class KeenSight(SkillTalent):
    name: Literal['Keen Sight']
    desc: str = (
        'Combat Action: When your active character is Tighnari, equip this '
        'card. After Tighnari equips this card, immediately use Vijnana-Phala '
        'Mine once. When your Tighnari, who has this card equipped, is '
        "affected by Vijnana Suffusion, the character's Charged Attack costs "
        '1 less Unaligned Element.'
    )
    version: Literal['3.6'] = '3.6'
    charactor_name: Literal['Tighnari'] = 'Tighnari'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 4,
    )
    skill: VijnanaPhalaMine = VijnanaPhalaMine()


# charactor base


class Tighnari(CharactorBase):
    name: Literal['Tighnari']
    version: Literal['3.6'] = '3.6'
    desc: str = '''"Verdant Strider" Tighnari'''
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | VijnanaPhalaMine | FashionersTanglevineShaft
    ] = []
    faction: List[FactionType] = [
        FactionType.SUMERU
    ]
    weapon_type: WeaponType = WeaponType.BOW
    talent: KeenSight | None = None

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Khanda Barrier-Buster',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            VijnanaPhalaMine(),
            FashionersTanglevineShaft()
        ]



from typing import Any, List, Literal

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillTalent
)


# Skills


class JadeScreen(ElementalSkillBase):
    name: Literal['Jade Screen'] = 'Jade Screen'
    desc: str = '''Deals 2 Geo DMG, creates 1 Jade Screen.'''
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_team_status(self.name),
        ]


class Starshatter(ElementalBurstBase):
    name: Literal['Starshatter'] = 'Starshatter'
    desc: str = (
        'Deals 6 Geo DMG. If Jade Screen is on the field, deals +2 DMG.'
    )
    damage: int = 6
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Check if Jade Screen exists. If so, change self.damage to 8.
        """
        status = match.player_tables[self.position.player_idx].team_status
        screen_exist: bool = False
        for s in status:
            if s.name == 'Jade Screen':
                screen_exist = True
                break
        if screen_exist:
            # temporary change damage to 8
            self.damage = 8
        ret = super().get_actions(match)
        self.damage = 6
        return ret


# Talents


class StrategicReserve(SkillTalent):
    name: Literal['Strategic Reserve']
    desc: str = (
        'Combat Action: When your active character is Ningguang, equip this '
        'card. After Ningguang equips this card, immediately use Jade Screen '
        'once. When your Ningguang, who has this card equipped, is on the '
        'field, Jade Screen will cause you to deal +1 Geo DMG.'
    )
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Ningguang'] = 'Ningguang'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 4,
    )
    skill: JadeScreen = JadeScreen()


# charactor base


class Ningguang(CharactorBase):
    name: Literal['Ningguang']
    version: Literal['3.3'] = '3.3'
    desc: str = '''"Eclipsing Star" Ningguang'''
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        ElementalNormalAttackBase | JadeScreen | Starshatter
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.CATALYST
    talent: StrategicReserve | None = None

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Sparkling Scatter',
                damage_type = self.element,
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            JadeScreen(),
            Starshatter(),
        ]



from typing import Any, List, Literal

from ....modifiable_values import DamageValue

from ....action import MakeDamageAction

from ....event import SkillEndEventArguments

from ....consts import (
    DamageElementalType, DamageType, ObjectPositionType, WeaponType
)

from ....struct import Cost
from .base import RoundEffectWeaponBase


class AquilaFavonia(RoundEffectWeaponBase):
    name: Literal['Aquila Favonia']
    desc: str = (
        'The character deals +1 DMG. After the opposing character uses a '
        'Skill: If the character with this attached is the active character, '
        'heal this character for 1 HP. (Max twice per Round) '
    )
    cost: Cost = Cost(same_dice_number = 3)
    version: Literal['3.3'] = '3.3'
    weapon_type: WeaponType = WeaponType.SWORD
    max_usage_per_round: int = 2

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self use elemental skill, create one corresponding dice
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = False,
            source_area = ObjectPositionType.CHARACTOR,
            target_area = ObjectPositionType.SKILL
        ):
            # not equipped or not opponent use skill
            return []
        if self.usage == 0:
            # no usage
            return []
        self.usage -= 1
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        return [MakeDamageAction(
            source_player_idx = self.position.player_idx,
            target_player_idx = self.position.player_idx,
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = charactor.position,
                    damage = -1,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = self.cost.copy()
                )
            ],
        )]


Swords = AquilaFavonia | AquilaFavonia



from .old_version import OldVersionCharactorStatus
from .system import SystemCharactorStatus
from .pyro_charactors import PyroCharactorStatus
from .dendro_charactors import DendroCharactorStatus
from .geo_charactors import GeoCharactorStatus
from .electro_charactors import ElectroCharactorStatus
from .anemo_charactors import AnemoCharactorStatus
from .hydro_charactors import HydroCharactorStatus
from .cryo_charactors import CryoCharactorStatus
from .foods import FoodStatus
from .artifacts import ArtifactCharactorStatus
from .weapons import WeaponCharactorStatus
from .event_cards import EventCardCharactorStatus


CharactorStatus = (
    PyroCharactorStatus | DendroCharactorStatus | GeoCharactorStatus
    | ElectroCharactorStatus | AnemoCharactorStatus | HydroCharactorStatus
    | CryoCharactorStatus
    | SystemCharactorStatus | FoodStatus | ArtifactCharactorStatus
    | WeaponCharactorStatus | EventCardCharactorStatus
    # finally old versions
    | OldVersionCharactorStatus
)



from typing import Any, List, Literal

from ...action import Actions, CreateObjectAction
from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillTalent
)


# Skills


class JumpyDumpty(ElementalSkillBase):
    name: Literal['Jumpy Dumpty'] = 'Jumpy Dumpty'
    desc: str = '''Deals 3 Pyro DMG. This character gains Explosive Spark.'''
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        status_usage = 1
        if self.is_talent_equipped(match):
            status_usage = 2
        return super().get_actions(match) + [
            self.create_charactor_status('Explosive Spark', {
                'usage': status_usage,
                'max_usage': status_usage,
            })
        ]


class SparksNSplash(ElementalBurstBase):
    name: Literal["Sparks 'n' Splash"] = "Sparks 'n' Splash"
    desc: str = (
        "Deals 3 Pyro DMG, creates 1 Sparks 'n' Splash at the opponent's "
        "play area."
    )
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Create teams status on opposite, so cannot call self.create_team_status
        """
        position = ObjectPosition(
            player_idx = 1 - self.position.player_idx,
            area = ObjectPositionType.TEAM_STATUS,
            id = -1
        )
        return super().get_actions(match) + [
            CreateObjectAction(
                object_name = self.name,
                object_position = position,
                object_arguments = {}
            )
        ]


# Talents


class PoundingSurprise(SkillTalent):
    name: Literal['Pounding Surprise']
    desc: str = (
        'Combat Action: When your active character is Klee, equip this card. '
        'After Klee equips this card, immediately use Jumpy Dumpty once. '
        'When your Klee, who has this card equipped, creates an Explosive '
        'Spark, its Usage(s) +1.'
    )
    version: Literal['3.4'] = '3.4'
    charactor_name: Literal['Klee'] = 'Klee'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3,
    )
    skill: JumpyDumpty = JumpyDumpty()


# charactor base


class Klee(CharactorBase):
    name: Literal['Klee']
    version: Literal['3.4'] = '3.4'
    desc: str = '''"Fleeing Sunlight" Klee'''
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        ElementalNormalAttackBase | JumpyDumpty | SparksNSplash
    ] = []
    faction: List[FactionType] = [
        FactionType.MONDSTADT
    ]
    weapon_type: WeaponType = WeaponType.CATALYST
    talent: PoundingSurprise | None = None

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Kaboom!',
                damage_type = self.element,
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            JumpyDumpty(),
            SparksNSplash(),
        ]



"""
Cards that can obtained in hands.
"""

from .event import EventCards
from .support import Supports
from .equipment.artifact import Artifacts
from .equipment.weapon import Weapons


Cards = Weapons | Artifacts | Supports | EventCards
